#!/usr/bin/env python

import os
import sys
import exifread
import datetime
from .geo import normalize_bearing
import uuid

sys.path.insert(0, os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..")))
import json


def eval_frac(value):
    if value.den == 0:
        return -1.0
    return float(value.num) / float(value.den)


def format_time(time_string):
    '''
    Format time string with invalid time elements in hours/minutes/seconds
    Format for the timestring needs to be "%Y_%m_%d_%H_%M_%S"

    e.g. 2014_03_31_24_10_11 => 2014_04_01_00_10_11
    '''
    subseconds = False
    data = time_string.split("_")
    hours, minutes, seconds = int(data[3]), int(data[4]), int(data[5])
    date = datetime.datetime.strptime("_".join(data[:3]), "%Y_%m_%d")
    subsec = 0.0
    if len(data) == 7:
        if float(data[6]) != 0:
            subsec = float(data[6]) / 10 ** len(data[6])
            subseconds = True
    date_time = date + \
                datetime.timedelta(hours=hours, minutes=minutes,
                                   seconds=seconds + subsec)
    return date_time, subseconds


def gps_to_decimal(values, reference):
    sign = 1 if reference in 'NE' else -1
    degrees = eval_frac(values[0])
    minutes = eval_frac(values[1])
    seconds = eval_frac(values[2])
    return sign * (degrees + minutes / 60 + seconds / 3600)


def exif_datetime_fields():
    '''
    Date time fields in EXIF
    '''
    return [["EXIF DateTimeOriginal",
             "Image DateTimeOriginal",
             "EXIF DateTimeDigitized",
             "Image DateTimeDigitized",
             "EXIF DateTime",
             "Image DateTime",
             "GPS GPSDate",
             "EXIF GPS GPSDate",
             "EXIF DateTimeModified"]]


def exif_gps_date_fields():
    '''
    Date fields in EXIF GPS
    '''
    return [["GPS GPSDate",
             "EXIF GPS GPSDate"]]


class ExifRead:
    '''
    EXIF class for reading exif from an image
    '''

    def __init__(self, filename, details=False):
        '''
        Initialize EXIF object with FILE as filename or fileobj
        '''
        self.filename = filename
        if type(filename) == str:
            with open(filename, 'rb') as fileobj:
                self.tags = exifread.process_file(fileobj, details=details)
        else:
            self.tags = exifread.process_file(filename, details=details)

    def _extract_alternative_fields(self, fields, default=None, field_type=float):
        '''
        Extract a value for a list of ordered fields.
        Return the value of the first existed field in the list
        '''
        for field in fields:
            if field in self.tags:
                if field_type is float:
                    value = eval_frac(self.tags[field].values[0])
                if field_type is str:
                    value = str(self.tags[field].values)
                if field_type is int:
                    value = int(self.tags[field].values[0])
                return value, field
        return default, None

    def exif_name(self):
        '''
        Name of file in the form {lat}_{lon}_{ca}_{datetime}_{filename}_{hash}
        '''
        lat, lon, ca, captured_at = self.exif_properties()

        filename = '{}_{}_{}_{}_{}'.format(
            lat, lon, ca, captured_at, uuid.uuid4())

        return filename

    def exif_properties(self):
        '''
        Gets {lat} {lon} {ca} {captured_at} as a tuple
        '''
        mapilio_description = json.loads(self.extract_image_description())

        lat = None
        lon = None
        ca = None
        captured_at = None

        if "MAPLatitude" in mapilio_description:
            lat = mapilio_description["MAPLatitude"]
        if "MAPLongitude" in mapilio_description:
            lon = mapilio_description["MAPLongitude"]
        if "MAPCompassHeading" in mapilio_description:
            if 'TrueHeading' in mapilio_description["MAPCompassHeading"]:
                ca = mapilio_description["MAPCompassHeading"]['TrueHeading']
        if "MAPCaptureTime" in mapilio_description:
            captured_at = datetime.datetime.strptime(
                mapilio_description["MAPCaptureTime"], "%Y_%m_%d_%H_%M_%S_%f").strftime("%Y-%m-%d-%H-%M-%S-%f")[:-3]

        return lat, lon, ca, captured_at

    def extract_image_history(self):
        field = ['Image Tag 0x9213']
        user_comment, _ = self._extract_alternative_fields(field, '{}', str)
        return user_comment

    def extract_altitude(self):
        '''
        Extract altitude
        '''
        altitude_ref = {
            0: 1,
            1: -1}
        fields = ['GPS GPSAltitude', 'EXIF GPS GPSAltitude']
        refs = ['GPS GPSAltitudeRef', 'EXIF GPS GPSAltitudeRef']
        altitude, _ = self._extract_alternative_fields(fields, 0, float)
        ref = 0 if not any([True for x in refs if x in self.tags]) else [
            self.tags[x].values for x in refs if x in self.tags][0][0]
        return altitude * altitude_ref[ref]

    def extract_capture_time(self):
        '''
        Extract capture time from EXIF
        return a datetime object
        TODO: handle GPS DateTime
        '''
        time_string = exif_datetime_fields()[0]
        capture_time, time_field = self._extract_alternative_fields(
            time_string, 0, str)
        if time_field in exif_gps_date_fields()[0]:
            capture_time = self.extract_gps_time()
            return capture_time

        if capture_time == 0:
            # try interpret the filename
            try:
                capture_time = datetime.datetime.strptime(os.path.basename(
                    self.filename)[:-4] + '000', '%Y_%m_%d_%H_%M_%S_%f')
            except:
                return None
        else:
            capture_time = capture_time.replace(" ", "_")
            capture_time = capture_time.replace(":", "_")
            capture_time = capture_time.replace(".", "_")
            capture_time = capture_time.replace("-", "_")
            capture_time = capture_time.replace(",", "_")
            capture_time = "_".join(
                [ts for ts in capture_time.split("_") if ts.isdigit()])
            capture_time, subseconds = format_time(capture_time)
            sub_sec = "0"
            if not subseconds:
                sub_sec = self.extract_subsec()
            capture_time = capture_time + \
                           datetime.timedelta(seconds=float("0." + sub_sec))

        return capture_time

    def extract_direction(self):
        '''
        Extract image direction (i.e. compass, heading, bearing)
        '''
        fields = ['GPS GPSImgDirection',
                  'EXIF GPS GPSImgDirection',
                  'GPS GPSTrack',
                  'EXIF GPS GPSTrack']
        direction, _ = self._extract_alternative_fields(fields)

        if direction is not None:
            direction = normalize_bearing(direction, check_hex=True)
        return direction

    def extract_dop(self):
        '''
        Extract dilution of precision
        '''
        fields = ['GPS GPSDOP', 'EXIF GPS GPSDOP']
        dop, _ = self._extract_alternative_fields(fields)
        return dop

    def extract_geo(self):
        '''
        Extract geo-related information from exif
        '''
        altitude = self.extract_altitude()
        dop = self.extract_dop()
        lon, lat = self.extract_lon_lat()
        d = {}
        if lon is not None and lat is not None:
            d['latitude'] = lat
            d['longitude'] = lon
        if altitude is not None:
            d['altitude'] = altitude
        if dop is not None:
            d['dop'] = dop

        if self.tags['EXIF FocalLength'] is not None:
            try:
                d['FocalLength'] = self.tags['EXIF FocalLength']
            except:
                assert "Focal Length is not in image Metadata"

        return d

    def extract_gps_time(self):
        '''
        Extract timestamp from GPS field.
        '''
        gps_date_field = "GPS GPSDate"
        gps_time_field = "GPS GPSTimeStamp"
        gps_time = 0
        if gps_date_field in self.tags and gps_time_field in self.tags:
            date = str(self.tags[gps_date_field].values).split(":")
            if int(date[0]) == 0 or int(date[1]) == 0 or int(date[2]) == 0:
                return None
            t = self.tags[gps_time_field]
            gps_time = datetime.datetime(
                year=int(date[0]),
                month=int(date[1]),
                day=int(date[2]),
                hour=int(eval_frac(t.values[0])),
                minute=int(eval_frac(t.values[1])),
                second=int(eval_frac(t.values[2])),
            )
            microseconds = datetime.timedelta(
                microseconds=int((eval_frac(t.values[2]) % 1) * 1e6))
            gps_time += microseconds
        return gps_time

    def extract_exif(self):
        '''
        Extract a list of exif infos
        '''
        width, height = self.extract_image_size()
        make, model = self.extract_make(), self.extract_model()
        orientation = self.extract_orientation()
        geo = self.extract_geo()
        capture = self.extract_capture_time()
        direction = self.extract_direction()
        d = {
            'width': width,
            'height': height,
            'orientation': orientation,
            'direction': direction,
            'make': make,
            'model': model,
            'capture_time': capture
        }
        d['gps'] = geo
        return d

    def extract_image_size(self):
        '''
        Extract image height and width
        '''
        width, _ = self._extract_alternative_fields(
            ['Image ImageWidth', 'EXIF ExifImageWidth'], -1, int)
        height, _ = self._extract_alternative_fields(
            ['Image ImageLength', 'EXIF ExifImageLength'], -1, int)
        return width, height

    def extract_image_description(self):
        '''
        Extract image description
        '''
        description, _ = self._extract_alternative_fields(
            ['Image ImageDescription'], "{}", str)
        return description

    def extract_lon_lat(self):
        if 'GPS GPSLatitude' in self.tags and 'GPS GPSLatitude' in self.tags:
            lat = gps_to_decimal(self.tags['GPS GPSLatitude'].values,
                                 self.tags['GPS GPSLatitudeRef'].values)
            lon = gps_to_decimal(self.tags['GPS GPSLongitude'].values,
                                 self.tags['GPS GPSLongitudeRef'].values)
        elif 'EXIF GPS GPSLatitude' in self.tags and 'EXIF GPS GPSLatitude' in self.tags:
            lat = gps_to_decimal(self.tags['EXIF GPS GPSLatitude'].values,
                                 self.tags['EXIF GPS GPSLatitudeRef'].values)
            lon = gps_to_decimal(self.tags['EXIF GPS GPSLongitude'].values,
                                 self.tags['EXIF GPS GPSLongitudeRef'].values)
        else:
            lon, lat = None, None
        return lon, lat

    def extract_make(self):
        '''
        Extract camera make
        '''
        fields = ['EXIF LensMake', 'Image Make']
        make, _ = self._extract_alternative_fields(
            fields, default='none', field_type=str)
        return make

    def extract_model(self):
        '''
        Extract camera model
        '''
        fields = ['EXIF LensModel', 'Image Model']
        model, _ = self._extract_alternative_fields(
            fields, default='none', field_type=str)
        return model

    def extract_firmware(self):
        '''
        Extract camera firmware (tag is called 'software' in EXIF)
        '''
        fields = ['Image Software']
        software, _ = self._extract_alternative_fields(
            fields, default="", field_type=str)
        return software

    def extract_orientation(self):
        '''
        Extract image orientation
        '''
        fields = ['Image Orientation']
        orientation, _ = self._extract_alternative_fields(
            fields, default=1, field_type=int)
        if orientation not in range(1, 9):
            return 1
        return orientation

    def extract_subsec(self):
        '''
        Extract microseconds
        '''
        fields = [
            'Image SubSecTimeOriginal',
            'EXIF SubSecTimeOriginal',
            'Image SubSecTimeDigitized',
            'EXIF SubSecTimeDigitized',
            'Image SubSecTime',
            'EXIF SubSecTime'
        ]
        sub_sec, _ = self._extract_alternative_fields(
            fields, default='', field_type=str)
        return sub_sec

    def fields_exist(self, fields):
        '''
        Check existence of a list fields in exif
        '''
        for rexif in fields:
            vflag = False
            for subrexif in rexif:
                if subrexif in self.tags:
                    vflag = True
            if not vflag:
                print("Missing required EXIF tag: {0} for image {1}".format(
                    rexif[0], self.filename))
                return False
        return True

    def mapilio_tag_exists(self):
        '''
        Check existence of required Mapilio tags
        '''
        description_tag = "Image ImageDescription"
        if description_tag not in self.tags:
            return False
        for requirement in ["MAPSequenceUUID", "MAPSettingsUserKey", "MAPCaptureTime", "MAPLongitude", "MAPLatitude"]:
            if requirement not in self.tags[description_tag].values or json.loads(self.tags[description_tag].values)[
                requirement] in ["", None, " "]:
                return False
        return True



import random
import re
import os
import numpy as np
from typing import Tuple
from addict import Dict
import string
import requests
import ast


class Generator:

    @staticmethod
    def unique_matchId_generator(letter_count: int = 12, digit_count: int = 8) -> str:
        str1 = ''.join((random.choice(string.ascii_letters) for x in range(letter_count)))
        str1 += ''.join((random.choice(string.digits) for x in range(digit_count)))

        sam_list = list(str1)  # it converts the string to list.
        random.shuffle(sam_list)  # It uses a random.shuffle() function to shuffle the string.
        final_string = ''.join(sam_list)
        return final_string

    @staticmethod
    def path_url_creator(**kwargs) -> Tuple[str, str]:
        """

        :param params.gui:
        :param params.cfg:
        :param params.splitData:
        :param params.index:
        :param params.directory:
        :return:
        """
        params = Dict(kwargs)
        i = params.index
        host = params.cfgImage.ip_remote if params.cfgImage.Remote else params.cfgImage.ip_local
        if params.gui:
            ip_path = re.compile("^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$")
            ip_check = bool(ip_path.match(params.directory))
            if ip_check:
                path = os.path.join(
                    "http://" + host + ":" + params.cfgImage.port + "/" + params.cfgImage.Directory + ":",
                    str(params.splitData[i]["dirname"]), params.splitData[i]["filename"], params.splitData[i]["imgname"])
            else:
                path = os.path.join(params.directory, str(params.splitData[i]["dirname"]),
                                    params.splitData[i]["filename"],
                                    params.splitData[i]["imgname"])
        else:
            # TODO will be dynamic connection for now test phase
            if params.cfgImage.device == "google":
                #     H:\IMAGERY\mapilio_dev\organization\arnavutkoy
                #     https://image.mapilio.com/h:/IMAGERY/mapilio_dev/organization/arnavutkoy/TqbHYpgLXzU7vHKWQDRVjQ.jpeg
                path = os.path.join("https://" + "image.mapilio.com", "h:", "IMAGERY", 'mapilio_dev', 'organization',
                                    'arnavutkoy', params.splitData[i]["imgname"])
            elif params.cfgImage.device == 'insta360':
                # http://10.10.5.49:3030/h:/BASAKSEHIR/ladybug_19027046_20200203_112314-000000/99.jpeg
                path = os.path.join("http://" + str(host) + ":" + str(params.cfgImage.port),
                                    "h:", params.splitData[i]["dirname"], params.splitData[i]["filepath"],
                                    params.splitData[i]["filename"])
            elif params.cfgImage.device == 'gopro':
                # ENERJISA
                path = os.path.join("http://" + str(host) + ":" + str(params.cfgImage.port),
                                    "h:", params.splitData[i]["dirname"], params.splitData[i]["filepath"],
                                    params.splitData[i]["filename"])
            elif params.cfgImage.device == 'ladybug':
                path = os.path.join(
                    "http://" + str(host) + ":" + str(params.cfgImage.port) + "/" + str(params.cfgImage.directory) + ":",
                    str(params.splitData[i]["dirname"]), params.splitData[i]["filename"], params.splitData[i]["imgname"])

        print('Image Path :', path)
        return host, path

    @staticmethod
    def data_separation(data: list, dividing_percentage: int) -> list:
        """
         Segmenting incoming data

        :param data: data to be processed
        :param dividing_percentage: percentage of data fragmented
        :return: predicted masks
        """
        percentage = int(len(data) / 100 * dividing_percentage)

        for i in range(0, len(data), percentage):
            # Create an index range for l of n items:
            yield data[i:i + percentage]

    @staticmethod
    def take_objects(matchedObjects: list, take: int = 2):

        for i in range(0, len(matchedObjects), take):
            if len(matchedObjects[i:i + take]) % take == 0:
                yield matchedObjects[i:i + take]

    @staticmethod
    def get_exif_information(img_info):
        """
        :param img_info: exif object
        :return: (lat, lon), orientation, (Height, Width), FocalLength, Altitude,
        """
        information = {}
        data = img_info.extract_exif()
        try:
            information["model"] = data["model"]
            information["coordx"] = data["gps"]["latitude"]
            information["coordy"] = data["gps"]["longitude"]
            information["width"] = data["width"]
            information["height"] = data["height"]

            # Focal Length
            fLen_obj = data["gps"]["FocalLength"]
            fLen_str = f"{fLen_obj}"
            fLen_arr = fLen_str.split("/")
            fLen = float(int(fLen_arr[0]) / int(fLen_arr[1]))
            information["FocalLength"] = fLen

            hor_width = data["height"] if data["orientation"] == 1 else data["width"]
            information["orientation"] = hor_width
            # Angle of View
            aFov = np.arctan(hor_width / (2 * fLen)) * (180 / np.pi)
            information["FoV"] = aFov
        except:
            raise Exception(f"Check the image Exif Data some missing values")

        for k, v in information.items():
            if information[k] is None:
                raise Exception(f"{k} is None")
            else:
                pass

        return information

    @staticmethod
    def get_config_response(configUrl: str) -> Dict:
        response = str(Dict(requests.get(configUrl).json()).config)
        pure_response = response.replace("\'", "\"")
        fix_dict = Dict(ast.literal_eval(pure_response))
        return fix_dict


import logging
from decimal import Decimal

NUMBER_TYPES = (int, float, Decimal)

__version__ = "1.1.7"
__name__ = "helper-mapilio"
__major__ = False
__minor__ = False
__patch__ = True

logger = logging.getLogger('helper')



import time
import warnings
from functools import wraps

from helper.rule import Utilities


class TimeLogger:
    def __init__(self):
        self._total_time_call_stack = [0]

    def timeLog(self):
        def _timeLog(fn):
            @wraps(fn)
            def wrapped_fn(*args, **kwargs):
                self._total_time_call_stack.append(0)

                start_time = time.time()

                try:
                    result = fn(*args, **kwargs)
                finally:
                    elapsed_time = time.time() - start_time

                    self._total_time_call_stack[-1] += elapsed_time

                    hours, rem = divmod(time.time() - start_time, 3600)
                    minutes, seconds = divmod(rem, 60)
                    elapsed_time = "{:0>2}:{:0>2}:{:05.2f}".format(int(hours), int(minutes), seconds)

                    # log the result
                    self.log_trigger(message={
                        'function_name': fn.__name__,
                        'total_time': elapsed_time,
                    })

                return result

            return wrapped_fn

        return _timeLog

    def log_trigger(self, message):
        self.timing_save_as_log_format(message)

    def timing_save_as_log_format(self, message):
        # TODO read PATH from config
        PATH = 'TimingResult.log'
        tLf = Utilities.file_exist_check_and_open(PATH)

        L = '[TimeTracker >] {function_name} {total_time}'.format(**message)
        tLf.writelines(L + '\n')
        tLf.close()


class Logger:

    @staticmethod
    def disable_warnings():
        # remove all the annoying warnings from tf v1.10 to v1.13
        import logging

        logging.getLogger('tensorflow').disabled = True

        def warn(*args, **kwargs):
            pass

        warnings.warn = warn


time_tracker = TimeLogger()
disable_warnings = Logger.disable_warnings()


class Handler:

    @staticmethod
    def error_information(db, exectype, value, tb):
        print(
            'database Connection',
            db)
        print(
            'My Error Information')
        print(
            'Type:', type(exectype).__name__)
        print(
            'Value:', value)
        print(
            'Traceback:', tb.format_exc().split('\n'))


"""
Mapilio Helper-tools is a python library for some image, prediction functions

Rule -> math functions
Exceptions -> handler errors or disable some warnings
Convertors -> tensorFaster_to_np, tensorMask_to_np, list_to_dict, url_to_image, Bunch
Generators ->path url creator, data separator, url to image
Logger -> time log, time save, file exist
"""


from .generator import Generator
from .logger import Logger, TimeLogger
from .handler import Handler
from .convertor import Convertor, JSON_ENCODER_FIX
from .rule import Utilities


from .util import __version__, __major__, __minor__, __patch__


import json

import numpy as np
import pandas as pd
from trianglesolver import solve, degree
from addict import Dict
import os
import csv
from urllib.parse import urlparse


class Utilities:

    @staticmethod
    def is_match(matched_objects: list) -> bool:
        """
        matched_objects matched objects twice
        :param matched_objects:
        :return: check out length bool type
        """
        if len(matched_objects) > 0:
            return True
        else:
            return False

    @staticmethod
    def is_between(t1max: bool, t2max: bool) -> bool:
        """
        the aim check between each detected thetas

        :param t1max: t1max is thetas max 1  between defined angle wide in config
        :param t2max: t2max is thetas max 1  between defined angle wide in config
        :return: check cross match as bool type
        """
        return True if (t1max and t2max) else False

    @staticmethod
    def is_right_or_left(**kwargs) -> bool:
        """
        Goal : detected objects must be same region each right or left.
        # theta1: float, heading1: float, theta2: float, heading2: float
        :param theta1: first object which seen by car angle
        :param heading1: moment car location angle according to north
        :param theta2: second object which seen by car angle
        :param heading2: moment car location angle according to north
        :return: gives answer as bool type objects left or right check
        """
        params = Dict(**kwargs)
        theta1 = params.theta1
        theta2 = params.theta2
        heading1 = params.heading1
        heading2 = params.heading2

        if theta1 <= heading1 and theta2 <= heading2:
            return True
        elif theta1 >= heading1 and theta2 >= heading2:
            return True
        else:
            return False

    @staticmethod
    def check_validity(a, b, c):
        """
         Triangle rule checks
        it's triangle edged as a, b, c
         5 < C / degree < 20 must be otherwise return false

        :param a: detected point and **first** car position edge
        :param b: detected point and **second** car position edge
        :param c: first and second car edge
        :return: if is not triangle and C degree over the defined variable return False.
        """

        a, b, c, A, B, C = solve(a, b, c)
        if (a + b <= c) or (a + c <= b) or (b + c <= a):
            if not 5 < (C / degree) < 20:
                return False
        return True

    @staticmethod
    def get_label_name_or_id(labelId: int, config: dict):
        """
         this function has two assignment

        :param labelId: if labelId is **array** return classname else
                        labelId is **string** return classname id
        :param config:
        :return: object's name or id
        """

        with open(config.labeljson, 'r') as f:
            labelsDict = json.load(f)
            if isinstance(labelId, np.ndarray):
                return labelsDict.get(str(labelId))
            elif isinstance(labelId, str):
                for id, name in labelsDict.items():
                    if name == labelId:
                        return id

    @staticmethod
    def file_exist_check_and_open(path: str):
        if os.path.exists(path):
            append_write = 'a'  # append if already exists
        else:
            append_write = 'w'  # make a new file if not

        return open(path, append_write)

    @staticmethod
    def file_check(file_id: str):
        if not os.path.exists(os.path.dirname("Exports")):
            try:
                os.makedirs("Exports")
            except OSError as err:
                print(err)
        os.makedirs(os.path.join("Exports", file_id, "ObjectsImage"))
        os.makedirs(os.path.join("Exports", file_id, "imgs"))

        header = ['ImagePath', 'ImageName']
        with open(os.path.join("Exports", file_id, "images_list.csv"), 'w', encoding='UTF8') as f:
            writer = csv.writer(f)
            writer.writerow(header)
            f.close()
        outPath = os.path.join("Exports", file_id)

        return outPath

    @staticmethod
    def config_reader(path: str) -> Dict:

        with open(path) as f:
            cfg = json.load(f)

        return Dict(cfg)

    @staticmethod
    def image_logger(url: str, file_id: str) -> bool:
        """

        Args:
            url: remote image url
            file_id: current processing task

        Returns: have not taken image false, otherwise true

        """
        df = pd.read_csv(os.path.join('Exports', file_id, 'images_list.csv'))

        url_check = df[df["ImagePath"] == url]
        if url_check.empty:
            url_parse = urlparse(url)
            image_name = os.path.basename(url_parse.path)
            df2 = {'ImagePath': url, 'ImageName': image_name}
            df = df.append(df2, ignore_index = True)
            df.to_csv(os.path.join('Exports', file_id, 'images_list.csv'), index=False)
            return False
        else:
            return True


import decimal
from json import JSONEncoder

import cv2
import numpy as np
import urllib.request
from operator import itemgetter
from itertools import groupby
from decimal import Decimal, ROUND_DOWN, getcontext


class Convertor:

    @staticmethod
    def url_to_image(url: str) -> np.ndarray:
        """
        # download the image, convert it to a NumPy array, and then read
        # it into OpenCV format
        :param url:
        :return:
        """

        resp = urllib.request.urlopen(url)
        image = np.asarray(bytearray(resp.read()), dtype="uint8")
        image = cv2.imdecode(image, cv2.IMREAD_COLOR)

        return image

    @staticmethod
    def tensorFaster_to_np(tensor):
        return tensor[0]["boxes"].cpu().numpy(), \
               tensor[0]["scores"].cpu().numpy(), \
               tensor[0]["labels"].cpu().numpy()

    @staticmethod
    def tensorMask_to_np(tensor):
        return tensor[0]["boxes"].detach().cpu().numpy(), \
               tensor[0]["scores"].detach().cpu().numpy(), \
               tensor[0]["labels"].detach().cpu().numpy(), \
               tensor[0]["masks"].squeeze().detach().cpu().numpy()

    @staticmethod
    def list_to_dict(data: list, config: dict) -> dict:
        """

        :param data:
        :param config:
        :return:
        """
        keys = config.columnKeys

        return dict(zip(keys, data))

    @staticmethod
    def url_to_auth_id(url):
        x = url.split("/")
        return x[5]

    @staticmethod
    class Bunch(object):
        def __init__(self, adict):
            self.__dict__.update(adict)

    @staticmethod
    def groupby_dict_in_list_key(arr, match):
        """

        Parameters
        ----------
        arr
        match

        Returns
        -------

        """
        arr = sorted(arr, key=itemgetter(match))  # very imported beacuse remove duplicate matched objects
        for key, value in groupby(arr, key=itemgetter(match)):
            yield value

    @staticmethod
    def decimal_fix(number):
        getcontext().rounding = ROUND_DOWN
        return Decimal(number).quantize(Decimal(10) ** -20)

    def get_rad(self, pitch, roll=0, yaw=0):
        return (self.deg_to_rad(float(pitch)),
                self.deg_to_rad(float(roll)),
                self.deg_to_rad(yaw))

    def deg_to_rad(self, deg):
        from math import pi
        return deg * pi / 180.0


class JSON_ENCODER_FIX(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, decimal.Decimal):
            return float(obj)

        return JSONEncoder.default(self, obj)



#!/usr/bin/env python
"""
Mapilio Helper
"""
from __future__ import (absolute_import, division, print_function)

import sys

from setuptools import find_packages, setup

if sys.version_info < (3, 5):
    raise RuntimeError(
        "Mapilio Helper tools supports Python 3.6 and above. "
    )

# This import must be below the above `sys.version_info` check,
# because the code being imported here is not compatible with the older
# versions of Python.
from increment_version import __version__
from helper.util import __name__

INSTALL_REQUIRES = [
    'trianglesolver==1.2',
    'opencv-python==4.5.1.48',
    'addict==2.4.0'
]

setup(
    name=__name__,
    version=__version__,
    description='Mapilio Helper Library',
    url='https://github.com/mapilio/helper.git',
    author='Mapilio - Ozcan Durak & M.Can VARER',
    author_email='ozcan@visiosoft.com.tr',
    license='licensed',
    packages=find_packages(),
    install_requires=INSTALL_REQUIRES,
    python_requires=">=3.6, <4",
    project_urls={  # Optional
            'Bug Reports': 'https://github.com/mapilio/helper/issues',
            'Say Thanks!': 'https://mapilio.com/#contact',
            'Source': 'https://github.com/mapilio/helper/',
        },
)


# -*- coding: utf-8 -*-

import datetime
import math

import datetime
import pytz

WGS84_a = 6378137.0
WGS84_b = 6356752.314245


def ecef_from_lla(lat, lon, alt):
    '''
    Compute ECEF XYZ from latitude, longitude and altitude.

    All using the WGS94 model.
    Altitude is the distance to the WGS94 ellipsoid.
    Check results here http://www.oc.nps.edu/oc2902w/coord/llhxyz.htm

    '''
    a2 = WGS84_a ** 2
    b2 = WGS84_b ** 2
    lat = math.radians(lat)
    lon = math.radians(lon)
    L = 1.0 / math.sqrt(a2 * math.cos(lat) ** 2 + b2 * math.sin(lat) ** 2)
    x = (a2 * L + alt) * math.cos(lat) * math.cos(lon)
    y = (a2 * L + alt) * math.cos(lat) * math.sin(lon)
    z = (b2 * L + alt) * math.sin(lat)
    return x, y, z


def gps_distance(latlon_1, latlon_2):
    '''
    Distance between two (lat,lon) pairs.

    >>> p1 = (42.1, -11.1)
    >>> p2 = (42.2, -11.3)
    >>> 19000 < gps_distance(p1, p2) < 20000
    True
    '''
    x1, y1, z1 = ecef_from_lla(latlon_1[0], latlon_1[1], 0.)
    x2, y2, z2 = ecef_from_lla(latlon_2[0], latlon_2[1], 0.)

    dis = math.sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2 + (z1 - z2) ** 2)

    return dis


def get_max_distance_from_start(latlon_track):
    '''
    Returns the radius of an entire GPS track. Used to calculate whether or not the entire sequence was just stationary video
    Takes a sequence of points as input
    '''
    latlon_list = []
    # Remove timestamps from list
    for idx, point in enumerate(latlon_track):
        lat = latlon_track[idx][1]
        lon = latlon_track[idx][2]
        alt = latlon_track[idx][3]
        latlon_list.append([lat, lon, alt])

    start_position = latlon_list[0]
    max_distance = 0
    for position in latlon_list:
        distance = gps_distance(start_position, position)
        if distance > max_distance:
            max_distance = distance
    return max_distance


def get_total_distance_traveled(latlon_track):
    '''
    Returns the total distance traveled of a GPS track. Used to calculate whether or not the entire sequence was just stationary video
    Takes a sequence of points as input
    '''
    latlon_list = []
    # Remove timestamps from list
    for idx, point in enumerate(latlon_track):
        lat = latlon_track[idx][1]
        lon = latlon_track[idx][2]
        alt = latlon_track[idx][3]
        latlon_list.append([lat, lon, alt])

    total_distance = 0
    last_position = latlon_list[0]
    for position in latlon_list:
        total_distance += gps_distance(last_position, position)
        last_position = position
    return total_distance


def gps_speed(distance, delta_t):
    # Most timestamps have 1 second resolution so change zeros in delta_t for
    # 0.5 so that we don't divide by zero
    delta_t_corrected = [0.5 if x == 0 else x for x in delta_t]
    speed = [distance / delta_t_corrected
             for distance, delta_t_corrected in zip(distance, delta_t_corrected)]
    return speed


def dms_to_decimal(degrees, minutes, seconds, hemisphere):
    '''
    Convert from degrees, minutes, seconds to decimal degrees.
    @author: mprins
    '''
    dms = float(degrees) + float(minutes) / 60 + float(seconds) / 3600
    if hemisphere in "WwSs":
        dms = -1 * dms

    return dms


def decimal_to_dms(value, precision):
    '''
    Convert decimal position to degrees, minutes, seconds in a fromat supported by EXIF
    '''
    deg = math.floor(value)
    min = math.floor((value - deg) * 60)
    sec = math.floor((value - deg - min / 60) * 3600 * precision)

    return ((deg, 1), (min, 1), (sec, precision))


def gpgga_to_dms(gpgga):
    '''
    Convert GPS coordinate in GPGGA format to degree/minute/second

    Reference: http://us.cactii.net/~bb/gps.py
    '''
    deg_min, dmin = gpgga.split('.')
    degrees = int(deg_min[:-2])
    minutes = float('%s.%s' % (deg_min[-2:], dmin))
    decimal = degrees + (minutes / 60)
    return decimal


def utc_to_localtime(utc_time):
    utc_offset_timedelta = datetime.datetime.utcnow() - datetime.datetime.now()
    return utc_time - utc_offset_timedelta


def compute_bearing(start_lat, start_lon, end_lat, end_lon):
    '''
    Get the compass bearing from start to end.

    Formula from
    http://www.movable-type.co.uk/scripts/latlong.html
    '''
    # make sure everything is in radians
    start_lat = math.radians(start_lat)
    start_lon = math.radians(start_lon)
    end_lat = math.radians(end_lat)
    end_lon = math.radians(end_lon)

    dLong = end_lon - start_lon

    dPhi = math.log(math.tan(end_lat / 2.0 + math.pi / 4.0) /
                    math.tan(start_lat / 2.0 + math.pi / 4.0))
    if abs(dLong) > math.pi:
        if dLong > 0.0:
            dLong = -(2.0 * math.pi - dLong)
        else:
            dLong = (2.0 * math.pi + dLong)

    y = math.sin(dLong) * math.cos(end_lat)
    x = math.cos(start_lat) * math.sin(end_lat) - \
        math.sin(start_lat) * math.cos(end_lat) * math.cos(dLong)
    bearing = (math.degrees(math.atan2(y, x)) + 360.0) % 360.0

    return bearing


def diff_bearing(b1, b2):
    '''
    Compute difference between two bearings
    '''
    d = abs(b2 - b1)
    d = 360 - d if d > 180 else d
    return d


def offset_bearing(bearing, offset):
    '''
    Add offset to bearing
    '''
    bearing = (bearing + offset) % 360
    return bearing


def normalize_bearing(bearing, check_hex=False):
    '''
    Normalize bearing and convert from hex if
    '''
    if bearing > 360 and check_hex:
        # fix negative value wrongly parsed in exifread
        # -360 degree -> 4294966935 when converting from hex
        bearing = bin(int(bearing))[2:]
        bearing = ''.join([str(int(int(a) == 0)) for a in bearing])
        bearing = -float(int(bearing, 2))
    bearing %= 360
    return bearing


def interpolate_lat_lon(points, t, max_dt=1):
    '''
    Return interpolated lat, lon and compass bearing for time t.

    Points is a list of tuples (time, lat, lon, elevation), t a datetime object.
    '''
    # find the enclosing points in sorted list
    if (t <= points[0][0]) or (t >= points[-1][0]):
        if t <= points[0][0]:
            dt = abs((points[0][0] - t).total_seconds())
        else:
            dt = (t - points[-1][0]).total_seconds()
        if dt > max_dt:
            raise ValueError(
                "time t not in scope of gpx file by {} seconds".format(dt))
        else:
            print(
                "time t not in scope of gpx file by {} seconds, extrapolating...".format(dt))

        if t < points[0][0]:
            before = points[0]
            after = points[1]
        else:
            before = points[-2]
            after = points[-1]
        bearing = compute_bearing(before[1], before[2], after[1], after[2])

        if t == points[0][0]:
            x = points[0]
            return (x[1], x[2], bearing, x[3])

        if t == points[-1][0]:
            x = points[-1]
            return (x[1], x[2], bearing, x[3])
    else:
        for i, point in enumerate(points):
            if t < point[0]:
                if i > 0:
                    before = points[i - 1]
                else:
                    before = points[i]
                after = points[i]
                break

    # weight based on time
    weight = (t - before[0]).total_seconds() / \
        (after[0] - before[0]).total_seconds()

    # simple linear interpolation in case points are not the same
    if before[1] == after[1]:
        lat = before[1]
    else:
        lat = before[1] - weight * before[1] + weight * after[1]

    if before[2] == after[2]:
        lon = before[2]
    else:
        lon = before[2] - weight * before[2] + weight * after[2]

    # camera angle
    bearing = compute_bearing(before[1], before[2], after[1], after[2])

    # altitude
    if before[3] is not None:
        ele = before[3] - weight * before[3] + weight * after[3]
    else:
        ele = None

    return lat, lon, bearing, ele


def write_gpx(filename, gps_trace):
    time_format = "%Y-%m-%dT%H:%M:%S.%f"
    gpx = "<gpx>" + "\n"
    gpx += "<trk>" + "\n"
    gpx += "<name>Mapillary GPX</name>" + "\n"
    gpx += "<trkseg>" + "\n"
    for point in gps_trace:
        lat = point[1]
        lon = point[2]
        if lat == 0 or lon == 0:
            continue
        time = datetime.datetime.strftime(point[0], time_format)[:-3]
        elevation = point[3] if len(point) > 3 else 0
        gpx += "<trkpt lat=\"" + \
            str(lat) + "\" lon=\"" + str(lon) + "\">" + "\n"
        gpx += "<ele>" + str(elevation) + "</ele>" + "\n"
        gpx += "<time>" + time + "</time>" + "\n"
        gpx += "</trkpt>" + "\n"
    gpx += "</trkseg>" + "\n"
    gpx += "</trk>" + "\n"
    gpx += "</gpx>" + "\n"
    with open(filename, "w") as fout:
        fout.write(gpx)


def get_timezone_and_utc_offset(lat, lon):
    #TODO Importing inside function because tzwhere is a temporary solution and dependency fails to install on windows
    from tzwhere import tzwhere

    tz = tzwhere.tzwhere(forceTZ=True) #TODO: This library takes 2 seconds to initialize. Should be done only once if used for many videos
    timezone_str = tz.tzNameAt(lat, lon)
    if timezone_str is not None:
        timezone = pytz.timezone(timezone_str)
        dt = datetime.datetime.utcnow()
        return [timezone_str, timezone.utcoffset(dt)]
    else:
        print("ERROR: Could not determine timezone")
        return [None, None]
